//+------------------------------------------------------------------+
//|                                                SOM_Net_Train.mqh |
//|                                            Copyright 2015, Urain |
//|                            https://login.mql5.com/ru/users/Urain |
//+------------------------------------------------------------------+
#property copyright "Copyright 2015, Urain"
#property link      "https://login.mql5.com/ru/users/Urain"
//+------------------------------------------------------------------+
//| #include <SOM_Net_Train.mqh>                                     |
//+------------------------------------------------------------------+
#include "SOM_Net_Data.mqh"
#include "PRNG_UD.mqh"
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class CSOM_Net_Train : public CSOM_Net_Data
  {
protected:
   int               m_iterations;
   double            m_map_radius;
   double            m_time_constant;
   C_PRNG_UD         rnd;
   int               iter;
   double            datavector[];
   double            learning_rate;
   double            m_initial_learning_rate;
   double            m_max_values[];
   double            m_min_values[];
   //---   
   void              NetInit();
   void              TrainInit();
   void              TrainIterations(int &p_iter);
   uint              gtc;
   double            time,res_time;
public:
                     CSOM_Net_Train(void):m_iterations(10),m_initial_learning_rate(0.1){};
                    ~CSOM_Net_Train(void){};
   //---   ,  
   void              InitParameters(int iterations,int xcells,int ycells);
   //---   
   void              Train();
   //---     
   void              SaveNet(string file_name);
   //---   (    CSOM_Net_Img)
   virtual void      Render(){};
   virtual void      ShowBMP(bool back){};
   virtual void      Progress(int p_iter);
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CSOM_Net_Train::Progress(int p_iter)
  {
   int timesek=((int)res_time%60);
   Comment(" ",p_iter,"  ",m_iterations," ",int(p_iter*100.0/m_iterations),
           "%   |     = ",(int)res_time/60," . ",(timesek<10?"0"+(string)timesek:(string)timesek)," .");
   Render(); ShowBMP(false);
   if(iter!=0)time=(GetTickCount()-gtc)/(1000.0*(iter/100.0));

   res_time=((m_iterations-p_iter)/100.)*time;
  };
//+------------------------------------------------------------------+
//|                                         |
//+------------------------------------------------------------------+
void CSOM_Net_Train::InitParameters(int iterations,int xcells,int ycells)
  {
   m_xsize=m_xcells=xcells;
   m_ysize=m_ycells=ycells;
//---        
   ArrayResize(m_som_nodes,xcells*ycells);
//---      
   m_iterations=iterations;
//---    
   if(m_xsize>m_ysize) {m_map_radius=m_xsize/2.0;}
   else {m_map_radius=m_ysize/2.0;}
//---
   m_time_constant=1.0*m_iterations/MathLog(m_map_radius);
   Print("m_time_constant=",m_time_constant);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CSOM_Net_Train::NetInit()
  {
   rnd.Init_Rand_Basis(m_total_patterns_sets);//     
   Print("total_training_sets = ",m_total_patterns_sets);
   learning_rate=m_initial_learning_rate;
   iter=0;
//---
   ArrayResize(datavector,m_dimension);
//---
   m_total_nodes=ArraySize(m_som_nodes);
//---
   Print("Dimension=",m_dimension);
//---        
   ArrayResize(m_max_values,m_dimension);
   ArrayResize(m_min_values,m_dimension);
//---
   for(int j=0; j<m_dimension; j++)
     {
      double maxv=m_patterns_sets_array[m_dimension+j];
      double minv=m_patterns_sets_array[m_dimension+j];
      for(int i=1; i<m_total_patterns_sets; i++)
        {
         double v=m_patterns_sets_array[m_dimension*i+j];
         if(v>maxv) {maxv=v;}
         if(v<minv) {minv=v;}
        }
      m_max_values[j]=maxv;
      m_min_values[j]=minv;
      Print(j,"m_min_value=",m_min_values[j],", m_max_value=",m_max_values[j]);
     }
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CSOM_Net_Train::TrainInit()
  {
   NetInit();
//---
   int cellwidth=m_xsize/m_xcells;
   int cellheight=m_ysize/m_ycells;
//---
   int indx=0;
//---  
   for(int i=0; i<m_xcells; i++)
     {
      for(int j=0; j<m_ycells; j++)
        {
         m_som_nodes[indx].SetDimension(m_dimension);
         m_som_nodes[indx].InitNode(i*cellwidth,j*cellheight,(i+1)*cellwidth,(j+1)*cellheight,m_min_values,m_max_values);
         indx++;
        }
     }
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CSOM_Net_Train::TrainIterations(int &p_iter)
  {
//---        
//int ind=int(m_total_training_sets*rand()/32768.);
   int ind=rnd.GetNext();
//---  datavector  
   for(int k=0; k<m_dimension; k++)
     {
      datavector[k]=m_patterns_sets_array[m_dimension*(ind)+k];
     }
//---    ,    datavector
   int winningnode=BestMatchingNode(datavector);
//---    
   double neighbourhood_radius=m_map_radius*exp(-1.0*p_iter/m_time_constant);
//---     
   double WS=neighbourhood_radius*neighbourhood_radius;
//---     
   for(int i=0; i<m_total_nodes; i++)
     {
      //---        m_WinningNodeIndex
      double DistToNodeSqr=(m_som_nodes[winningnode].X()-m_som_nodes[i].X())*(m_som_nodes[winningnode].X()-m_som_nodes[i].X())
                           +(m_som_nodes[winningnode].Y()-m_som_nodes[i].Y())*(m_som_nodes[winningnode].Y()-m_som_nodes[i].Y());
      //---    ,   
      if(DistToNodeSqr<WS)
        {
         //---     
         double influence=exp(-DistToNodeSqr/(2*WS));
         //---       
         m_som_nodes[i].AdjustWeights(datavector,learning_rate,influence);
        }
     }
//---  ( iter)   
   learning_rate=m_initial_learning_rate*exp(-1.0*p_iter/m_iterations);
   iter++;
  }
//+------------------------------------------------------------------+
//|                                              |
//+------------------------------------------------------------------+
void CSOM_Net_Train::Train()
  {
   TrainInit();
   time=0; res_time=0; gtc=GetTickCount();
//---
   do
     {
      TrainIterations(iter);//---   
      //---      
      //---      N- 
      if(iter%100==0)
        {
         Progress(iter);
/*
         //---      .
         int WIDTH=1465,HEIGHT=855,itername=(int(iter*100.0/m_iterations));
         string name="SOM\\ScreenShot\\Train SOM "+(itername<10?"0"+(string)itername:(string)itername)+".bmp";
         if(itername==100)name="SOM\\ScreenShot\\Train SOM 999.bmp";         
         ChartScreenShot(0,name,WIDTH,HEIGHT,ALIGN_LEFT);
*/
        }
      if(IsStopped())break;
     }
   //---     ,       
   while(iter<m_iterations);
   Comment("");
   Print(" .");
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CSOM_Net_Train::SaveNet(string file_name)
  {
   int size_net=ArraySize(m_som_nodes);
   double save_node[];
   if(size_net>0)
     {
      int dimension_node=m_som_nodes[0].GetDimension();
      int shift_index=6+dimension_node;
      double save_net[];
      int total_save_net=ArrayResize(save_net,DIMENSION_HEADER+shift_index*size_net);
      //--- Save Hendle Net
      int cnt=0;
      save_net[cnt]=shift_index;  cnt++;
      save_net[cnt]=m_xcells;     cnt++;
      save_net[cnt]=m_ycells;     cnt++;
      save_net[cnt]=m_xsize;      cnt++;
      save_net[cnt]=m_ysize;      cnt++;
      //--- Save nodes
      for(int i=0;i<size_net;i++)
        {
         m_som_nodes[i].SaveNode(save_node);
         for(int j=0;j<shift_index;j++)
           {
            save_net[cnt]=save_node[j];  cnt++;
           }
        }
      //---      double
      /*
      double save[]; uchar txtsave[]; int count_save=0;
      for(int i=0;i<dimension_node;i++)
        {
         uchar txtarr[];
         int count_char=StringToCharArray(m_som_titles[i]+";",txtarr);
         count_char=ArrayResize(txtarr,count_char-1);
         int size=ArraySize(txtsave);
         count_save=ArrayResize(txtsave,size+count_char);
         for(int j=0;j<count_char;j++) {txtsave[size+j]=txtarr[j];}
        }
      //---
      int count_double=(count_save%8==0?count_save/8:count_save/8+1);
      int size_txt=ArrayResize(txtsave,count_double*8);
      for(int i=size_txt-1;i>=count_save;i--)txtsave[i]=0;
      ArrayResize(c_double,count_double); ArrayResize(c_uchar,count_double);
      int scnt=0;
      for(int i=0;i<count_double;i++)
        {
         for(int j=0;j<8;j++) {c_uchar[i].c[j]=txtsave[scnt]; scnt++;}
         c_double[i]=(Conversion_Double)c_uchar[i];
        }
      //---
      total_save_net=ArrayResize(save_net,ArraySize(save_net)+count_double);
      for(int i=0;i<count_double;i++)
        {
         save_net[cnt]=c_double[i].d; cnt++;
        }
      */
      //---    
      ResetLastError();
      int hndl=FileOpen(file_name+".somnet",FILE_BIN|FILE_WRITE);
      if(hndl!=INVALID_HANDLE)
        {
         FileWriteInteger(hndl, cnt);
         FileWriteArray(hndl, save_net);
         FileWriteInteger(hndl, dimension_node);
         
         for(int i=0;i<dimension_node;i++)
         {
           FileWriteInteger(hndl, StringLen(m_som_titles[i]));
           FileWriteString(hndl, m_som_titles[i]);
         }
         
         FileClose(hndl);
        }
      else Print("Error save file "+file_name+".somnet , err="+(string)GetLastError());
     }
  };
//+------------------------------------------------------------------+
